/******************************************************************************/
/* SlunkCrypt, by LoRd_MuldeR <MuldeR2@GMX.de>                                */
/* This work has been released under the CC0 1.0 Universal license!           */
/******************************************************************************/

/* Internal */
#include "slunkcrypt.h"
#include "compiler.h"
#include "keygen.h"
#include "thread.h"
#include "version.h"

/* CRT */
#include <string.h>
#include <limits.h>

/* Version */
const uint16_t SLUNKCRYPT_VERSION_MAJOR = LIB_VERSION_MAJOR;
const uint16_t SLUNKCRYPT_VERSION_MINOR = LIB_VERSION_MINOR;
const uint16_t SLUNKCRYPT_VERSION_PATCH = LIB_VERSION_PATCH;
const char *const SLUNKCRYPT_BUILD = __DATE__ ", " __TIME__;

/* Utilities */
#define BOOLIFY(X) (!!(X))
#define THREAD_COUNT(X) (((X)->thread_pool) ? slunkcrypt_thrdpl_count((X)->thread_pool) : 1U)

/* Configuration */
const int SLUNKCRYPT_HAVE_THREADS = BOOLIFY(MAX_THREADS > 1U);

// ==========================================================================
// Data structures
// ==========================================================================

typedef struct
{
	uint32_t x, y, z, w, v, d;
}
rand_state_t;

typedef struct
{
	int reverse_mode;
	const uint8_t (*wheel)[256U];
	uint32_t counter;
	size_t index_off;
	rand_state_t random;
}
thread_state_t;

typedef struct
{
	uint8_t wheel[256U][256U];
	thread_state_t thread_data[MAX_THREADS];
}
crypt_data_t;

typedef struct
{
	int legacy_init;
	thrdpl_t *thread_pool;
	crypt_data_t data;
}
crypt_state_t;

// ==========================================================================
// Abort flag
// ==========================================================================

volatile int g_slunkcrypt_abort_flag = 0;

#define CHECK_ABORTED() do \
{ \
	if (g_slunkcrypt_abort_flag) \
	{ \
		goto aborted; \
	} \
} \
while (0)

// ==========================================================================
// Byte access (endianness agnostic)
// ==========================================================================

static INLINE uint32_t lower_u64(const uint64_t value)
{
	return (uint32_t)(value & 0xFFFFFFFF);
}

static INLINE uint32_t upper_u64(const uint64_t value)
{
	return (uint32_t)(value >> 32U);
}

// ==========================================================================
// Deterministic random bit generator
// ==========================================================================

static INLINE void random_init(rand_state_t *const state, const keydata_t *const key)
{
	slunkcrypt_bzero(state, sizeof(rand_state_t));
	state->x = lower_u64(key->a);
	state->y = upper_u64(key->a);
	state->z = lower_u64(key->b);
	state->w = upper_u64(key->b);
	state->v = lower_u64(key->c);
	state->d = upper_u64(key->c);
}

static INLINE uint32_t random_next(rand_state_t *const state)
{
	const uint32_t t = state->x ^ (state->x >> 2);
	state->x = state->y;
	state->y = state->z;
	state->z = state->w;
	state->w = state->v;
	state->v ^= (state->v << 4) ^ t ^ (t << 1);
	return (state->d += 0x000587C5) + state->v;
}

static INLINE void random_skip(rand_state_t *const state, const size_t skip_count)
{
	size_t i;
	for (i = 0U; i < skip_count; ++i)
	{
		UNUSED /*volatile*/ uint32_t q = random_next(state);
	}
}

static INLINE void random_seed(rand_state_t *const state, uint64_t salt, const uint16_t pepper, const uint8_t *const passwd, const size_t passwd_len, const int legacy)
{
	keydata_t key;
	do
	{
		slunkcrypt_keygen(&key, salt++, pepper, passwd, passwd_len, legacy ? 99971U : 16487L);
		random_init(state, &key);
		slunkcrypt_bzero(&key, sizeof(keydata_t));
	}
	while (!(state->x || state->y || state->z || state->w || state->v));
	random_skip(state, 97U);
}

// ==========================================================================
// Initialization
// ==========================================================================

static const uint16_t INITIALIZER[256U] =
{
	0x5C33, 0x5C4B, 0x5C6F, 0x5D0B, 0x5D47, 0x5DD7, 0x5E13, 0x5E8B, 0x5EAF, 0x5F57, 0x5FCF, 0x6053, 0x62DB, 0x62FF, 0x63EF, 0x642B,
	0x64C7, 0x64F7, 0x653F, 0x6593, 0x65F3, 0x675B, 0x6803, 0x6917, 0x692F, 0x69E3, 0x6A67, 0x6AA3, 0x6B87, 0x6C77, 0x6C9B, 0x6CB3,
	0x6D73, 0x6E03, 0x6E93, 0x6E9F, 0x6F53, 0x6FBF, 0x6FE3, 0x7133, 0x71C3, 0x723B, 0x729B, 0x72D7, 0x732B, 0x737F, 0x73DF, 0x743F,
	0x74B7, 0x75FB, 0x7673, 0x768B, 0x7703, 0x774B, 0x7853, 0x79A3, 0x7A1B, 0x7B3B, 0x7B5F, 0x7B77, 0x7C67, 0x7D03, 0x7DB7, 0x7EFB,
	0x7F5B, 0x804B, 0x80DB, 0x8153, 0x8243, 0x8357, 0x836F, 0x83E7, 0x8447, 0x860F, 0x863F, 0x8717, 0x876B, 0x878F, 0x8957, 0x8A0B,
	0x8A23, 0x8C27, 0x8C7B, 0x8CF3, 0x8DA7, 0x8DCB, 0x8E1F, 0x8E73, 0x8F4B, 0x9017, 0x903B, 0x9257, 0x92AB, 0x92E7, 0x93A7, 0x95B7,
	0x96CB, 0x96EF, 0x975B, 0x977F, 0x97D3, 0x98C3, 0x98CF, 0x999B, 0x99FB, 0x9A4F, 0x9B63, 0x9B93, 0x9B9F, 0x9C2F, 0x9CBF, 0x9E33,
	0x9F23, 0x9F53, 0x9F77, 0x9FB3, 0xA1AB, 0xA223, 0xA22F, 0xA247, 0xA26B, 0xA31F, 0xA373, 0xA397, 0xA3C7, 0xA427, 0xA4C3, 0xA53B,
	0xA577, 0xA5CB, 0xA6BB, 0xA70F, 0xA757, 0xA8D7, 0xA98B, 0xA9A3, 0xAA57, 0xAB0B, 0xABFB, 0xAC7F, 0xACEB, 0xADE7, 0xADFF, 0xAE8F,
	0xAE9B, 0xAF13, 0xAF43, 0xAF8B, 0xAFA3, 0xB01B, 0xB1BF, 0xB28B, 0xB33F, 0xB477, 0xB5DF, 0xB633, 0xB657, 0xB747, 0xB813, 0xB867,
	0xB8EB, 0xB993, 0xBA17, 0xBACB, 0xBB5B, 0xBD17, 0xBD5F, 0xBDB3, 0xBF33, 0xBF7B, 0xBFCF, 0xC047, 0xC0D7, 0xC15B, 0xC317, 0xC3E3,
	0xC51B, 0xC767, 0xC803, 0xC857, 0xC893, 0xC8CF, 0xC98F, 0xC9E3, 0xCA73, 0xCA7F, 0xCBFF, 0xCC9B, 0xCEB7, 0xCF0B, 0xCF9B, 0xD013,
	0xD1C3, 0xD1FF, 0xD217, 0xD2E3, 0xD40F, 0xD523, 0xD6AF, 0xD727, 0xD73F, 0xD77B, 0xD8BF, 0xD8FB, 0xD943, 0xD9EB, 0xDA9F, 0xDAC3,
	0xDAE7, 0xDDE7, 0xDEA7, 0xDF1F, 0xDF37, 0xDFC7, 0xE09F, 0xE0F3, 0xE177, 0xE1CB, 0xE22B, 0xE243, 0xE2BB, 0xE2D3, 0xE423, 0xE537,
	0xE567, 0xE61B, 0xE657, 0xE6B7, 0xE837, 0xE963, 0xEC7B, 0xECE7, 0xED6B, 0xEDE3, 0xEF9F, 0xEFC3, 0xF06B, 0xF107, 0xF20F, 0xF257,
	0xF35F, 0xF42B, 0xF4AF, 0xF5DB, 0xF743, 0xF7D3, 0xF863, 0xF86F, 0xF8E7, 0xFA07, 0xFA13, 0xFB1B, 0xFB3F, 0xFCFB, 0xFE27, 0xFE63
};

static int initialize_state(crypt_data_t *const data, const size_t thread_count, const uint64_t nonce, const uint8_t *const passwd, const size_t passwd_len, const int mode, const int legacy)
{
	uint8_t temp[256U][256U];
	size_t r, i;
	const int reverse_mode = BOOLIFY(mode);

	/* initialize state */
	slunkcrypt_bzero(data, sizeof(crypt_data_t));

	/* initialize counter */
	random_seed(&data->thread_data[0].random, nonce, legacy ? ((uint16_t)(-1)) : 0xFE7B, passwd, passwd_len, legacy);
	data->thread_data[0].counter = random_next(&data->thread_data[0].random);

	/* set up the wheel permutations */
	for (r = 0U; r < 256U; ++r)
	{
		random_seed(&data->thread_data[0].random, nonce, legacy ? ((uint16_t)r) : INITIALIZER[r], passwd, passwd_len, legacy);
		for (i = 0U; i < 256U; ++i)
		{
			const size_t j = random_next(&data->thread_data[0].random) % (i + 1U);
			if (j != i)
			{
				data->wheel[r][i] = data->wheel[r][j];
			}
			data->wheel[r][j] = (uint8_t)i;
		}
		CHECK_ABORTED();
	}

	/* reverse the wheels, if requested */
	if (reverse_mode)
	{
		for (r = 0U; r < 256U; ++r)
		{
			for (i = 0U; i < 256U; ++i)
			{
				temp[r][data->wheel[r][i]] = (uint8_t)i;
			}
		}
		for (r = 0U; r < 256U; ++r)
		{
			memcpy(data->wheel[255U - r], temp[r], 256U);
		}
		slunkcrypt_bzero(temp, sizeof(temp));
		CHECK_ABORTED();
	}

	/* initialize thread state */
	data->thread_data[0].reverse_mode = reverse_mode;
	data->thread_data[0].wheel = (const uint8_t(*)[256]) data->wheel;
	data->thread_data[0].index_off = 0U;
	random_seed(&data->thread_data[0].random, nonce, legacy ? 256U : 0xFEF3, passwd, passwd_len, legacy);
	for (i = 1U; i < thread_count; ++i)
	{
		data->thread_data[i].reverse_mode = data->thread_data[0].reverse_mode;
		data->thread_data[i].wheel = data->thread_data[0].wheel;
		data->thread_data[i].counter = data->thread_data[0].counter + ((uint32_t)i);
		data->thread_data[i].index_off = data->thread_data[i - 1U].index_off + 1U;
		memcpy(&data->thread_data[i].random, &data->thread_data[0].random, sizeof(rand_state_t));
		random_skip(&data->thread_data[i].random, i * 63U);
		CHECK_ABORTED();
	}

	return SLUNKCRYPT_SUCCESS;

	/* aborted */
aborted:
	slunkcrypt_bzero(data, sizeof(crypt_data_t));
	return SLUNKCRYPT_ABORTED;
}

// ==========================================================================
// Encrypt / Decrypt
// ==========================================================================

static INLINE void update_offset(uint8_t *const offset, uint32_t seed, rand_state_t *const state, const int reverse)
{
	size_t i;
	for (i = 0U; i < 256U; ++i, seed >>= CHAR_BIT)
	{
		if (i && (!(i & 3U)))
		{
			seed = random_next(state);
		}
		offset[reverse ? (255U - i) : i] = (uint8_t)seed;
	}
}

static INLINE uint8_t process_next_symbol(thread_state_t *const state, uint8_t value)
{
	uint8_t offset[256U];
	size_t i;
	update_offset(offset, state->counter, &state->random, state->reverse_mode);
	for (i = 0U; i < 256U; ++i)
	{
		value = (state->wheel[i][(value + offset[i]) & 0xFF] - offset[i]) & 0xFF;
	}
	return value;
}

// ==========================================================================
// Thread entry point
// ==========================================================================

static INLINE void update_index(thread_state_t *const state, const size_t thread_count, const size_t length)
{
	const size_t remaining = thread_count - (length % thread_count);
	if (remaining != thread_count)
	{
		state->index_off = (state->index_off + remaining) % thread_count;
	}
}

static void thread_worker(const size_t thread_count, void *const context, uint8_t *const buffer, const size_t length)
{
	thread_state_t *const state = (thread_state_t*) context;
	size_t i;

	for (i = state->index_off; i < length; i += thread_count)
	{
		buffer[i] = process_next_symbol(state, buffer[i]);
		state->counter += (uint32_t)thread_count;
		random_skip(&state->random, 63U * (thread_count - 1U));
	}

	update_index(state, thread_count, length);
}

// ==========================================================================
// Public API
// ==========================================================================

int slunkcrypt_generate_nonce(uint64_t *const nonce)
{
	if (!nonce)
	{
		return SLUNKCRYPT_FAILURE;
	}
	do
	{
		if (slunkcrypt_random_bytes((uint8_t*)nonce, sizeof(uint64_t)) != sizeof(uint64_t))
		{
			return SLUNKCRYPT_FAILURE;
		}
	}
	while (!(*nonce));
	return SLUNKCRYPT_SUCCESS;
}

slunkcrypt_t slunkcrypt_alloc(const uint64_t nonce, const uint8_t *const passwd, const size_t passwd_len, const int mode)
{
	slunkparam_t param = { SLUNKCRYPT_PARAM_VERSION, 0U, 1 };
	return slunkcrypt_alloc_ext(nonce, passwd, passwd_len, mode, &param);
}

slunkcrypt_t slunkcrypt_alloc_ext(const uint64_t nonce, const uint8_t *const passwd, const size_t passwd_len, const int mode, const slunkparam_t *const param)
{
	crypt_state_t* state = NULL;

	if ((!passwd) || (passwd_len < SLUNKCRYPT_PWDLEN_MIN) || (passwd_len > SLUNKCRYPT_PWDLEN_MAX) ||
		(mode < SLUNKCRYPT_ENCRYPT) || (mode > SLUNKCRYPT_DECRYPT) || (!param) || (param->version == 0U) || (param->version > SLUNKCRYPT_PARAM_VERSION))
	{
		return SLUNKCRYPT_NULL;
	}

	if (!(state = (crypt_state_t*)malloc(sizeof(crypt_state_t))))
	{
		return SLUNKCRYPT_NULL;
	}

	if ((state->thread_pool = slunkcrypt_thrdpl_create(param->thread_count, thread_worker)))
	{
		const size_t thread_count = slunkcrypt_thrdpl_count(state->thread_pool);
		size_t i;
		for (i = 0U; i < thread_count; ++i)
		{
			slunkcrypt_thrdpl_init(state->thread_pool, i, &state->data.thread_data[i]);
		}
	}

	state->legacy_init = (param->version > 1U) ? BOOLIFY(param->legacy_init) : 1;

	if (initialize_state(&state->data, THREAD_COUNT(state), nonce, passwd, passwd_len, mode, state->legacy_init) == SLUNKCRYPT_SUCCESS)
	{
		return (slunkcrypt_t)state;
	}

	slunkcrypt_free((slunkcrypt_t)state);
	return SLUNKCRYPT_NULL;
}

int slunkcrypt_reset(const slunkcrypt_t context, const uint64_t nonce, const uint8_t *const passwd, const size_t passwd_len, const int mode)
{
	crypt_state_t *const state = (crypt_state_t*) context;
	int result = SLUNKCRYPT_FAILURE;

	if ((!state) || (!passwd) || (passwd_len < SLUNKCRYPT_PWDLEN_MIN) || (passwd_len > SLUNKCRYPT_PWDLEN_MAX) || (mode < SLUNKCRYPT_ENCRYPT) || (mode > SLUNKCRYPT_DECRYPT))
	{
		return SLUNKCRYPT_FAILURE;
	}

	if ((result = initialize_state(&state->data, THREAD_COUNT(state), nonce, passwd, passwd_len, mode, state->legacy_init)) != SLUNKCRYPT_SUCCESS)
	{
		slunkcrypt_bzero(&state->data, sizeof(crypt_data_t));
	}

	return result;
}

int slunkcrypt_process(const slunkcrypt_t context, const uint8_t *const input, uint8_t *const output, size_t length)
{
	crypt_state_t *const state = (crypt_state_t*)context;
	if (!state)
	{
		return SLUNKCRYPT_FAILURE;
	}

	if (length > 0U)
	{
		memcpy(output, input, length);
		return slunkcrypt_inplace(context, output, length);
	}

	return SLUNKCRYPT_SUCCESS;
}

int slunkcrypt_inplace(const slunkcrypt_t context, uint8_t *const buffer, size_t length)
{
	crypt_state_t *const state = (crypt_state_t*)context;
	if (!state)
	{
		return SLUNKCRYPT_FAILURE;
	}

	if (length > 0U)
	{
		if (THREAD_COUNT(state) > 1U)
		{
			slunkcrypt_thrdpl_exec(state->thread_pool, buffer, length);
		}
		else
		{
			thread_worker(1U, &state->data.thread_data[0U], buffer, length);
		}
	}

	CHECK_ABORTED();
	return SLUNKCRYPT_SUCCESS;

aborted:
	slunkcrypt_bzero(buffer, length);
	slunkcrypt_bzero(&state->data, sizeof(crypt_data_t));
	return SLUNKCRYPT_ABORTED;
}

void slunkcrypt_free(const slunkcrypt_t context)
{
	crypt_state_t *const state = (crypt_state_t*) context;
	if (state)
	{
		if (state->thread_pool)
		{
			slunkcrypt_thrdpl_destroy(state->thread_pool);
		}
		slunkcrypt_bzero(state, sizeof(crypt_state_t));
		free(state);
	}
}
